import yaml
import argparse
import sys
import os.path
import glob
import shutil

data_dir = None

def error(msg):
    sys.stderr.write(msg + '\n')
    sys.exit(1)


class Result(object):
    def __init__(self):
        self.top_template = None
        self.mt_config = None


def parse(filename):
    d = yaml.load(open(filename))
    if 'wrtd-board' not in d:
        error('YAML file must have the wrtd-board key')
    return d['wrtd-board']


def build_mt_config(desc):
    appid = desc['appid']
    cpus = desc['cpus']
    ncpus = len([ x for x in cpus if isinstance(x, int) ])
    res = """
  constant c_mt_config : t_mt_config :=
    (
      app_id                        => x"{:08x}",
      cpu_count                     => {},
      cpu_config                    => (""".format(
          appid, ncpus)
    # Be sure that cpus are defined from 0 to N-1.
    for i in range(ncpus):
        cpu = desc['cpus'].get(i, None)
        if cpu == None:
            error('No definition for cpu {}'.format(i))
        res += """
        {}                           => (
          memsize                   => {},
          hmq_config                => (""".format(
                  i, cpu['memsize'] * 1024 // 4)

        hmq = cpu.get('hmq', None)
        if hmq is not None:
            res += """
            slot_count              => 1,
            slot_config             => (
              0                     => (
                entries_bits        => {},
                width_bits          => 7,
                header_bits         => 2,
                endpoint_id         => x"0000_0000",
                enable_config_space => FALSE),""".format(
                  hmq['log_entries'])
        res += """
              others                => c_DUMMY_MT_MQUEUE_SLOT)),
          rmq_config                => ("""

        rmq = cpu.get('rmq', None)
        if rmq is not None:
            res += """
            slot_count              => 1,
            slot_config             => (
              0                     => (
                entries_bits        => {},
                width_bits          => 7,
                header_bits         => 4,
                endpoint_id         => x"0000_0000",
                enable_config_space => TRUE),""".format(
                    rmq['log_entries'])
        res += """
              others                => c_DUMMY_MT_MQUEUE_SLOT))),"""
    res += """
        others                      => (
          0, c_MT_DEFAULT_MQUEUE_CONFIG, c_MT_DEFAULT_MQUEUE_CONFIG)),
      shared_mem_size               => {}
      );""".format(cpus['shared_memsize'] * 1024 // 4)

    return res

def board_svec(res, desc, slots):
    if len(slots) > 2:
        error('SVEC can have only 2 slots')
    for k in slots.keys():
        if k not in [0, 1]:
            error('incorrect slot id {} for SVEC'.format(k))

boards = { 'svec': board_svec }


def maybe_mkdir(path):
    try:
        os.mkdir(path)
    except FileExistsError:
        pass

def compute_fetchto(relfile):
    fetchto_dir = os.path.join(data_dir, "..", "dependencies")
    if os.path.isabs(data_dir):
        return os.path.normpath(fetchto_dir)
    else:
        return os.path.relpath(fetchto_dir, os.path.dirname(relfile))


def generate_hdl(res, board, slots):
    # Build ports
    ports = ""
    for k, v in slots.items():
        tmpl = open(os.path.join(data_dir, v, 'top.vhd')).read()
        ports += tmpl.format(n=k)
        ports += '\n'
    res.ports = ports

    # Build instances
    insts = ""
    for k, v in slots.items():
        tmpl = open(os.path.join(data_dir, v, 'inst.vhd')).read()
        insts += tmpl.format(n=k)
        insts += '\n'
    res.instances = insts

    # Top vhdl file
    maybe_mkdir("hdl")
    maybe_mkdir(os.path.join("hdl", "top"))
    maybe_mkdir(os.path.join("hdl", "top", res.name))
    top_filename = os.path.join("hdl", "top", res.name,
                                "{}_top.vhd".format(res.name))
    print("Writing {}".format(top_filename))
    top_template = open(os.path.join(data_dir, board, 'top.vhd')).read()

    top = open(top_filename, "w")
    top.write(top_template.format(name=res.name,
                                  ports=res.ports,
                                  mt_config=res.mt_config,
                                  instances=res.instances))
    top.close()

    # Top Manifest
    manifest_filename = os.path.join("hdl", "top", res.name, "Manifest.py")
    manifest_template = open(
        os.path.join(data_dir, board, 'Manifest-top.py')).read()
    modules_set = {v for v in slots.values()}
    modules = ""
    for m in modules_set:
        repo = open(os.path.join(data_dir, m, 'repo.txt')).read().rstrip()
        modules += '        "{}",\n'.format(repo)

    print("Writing {}".format(manifest_filename))
    manifest = open(manifest_filename, "w")
    dependencies = compute_fetchto(manifest_filename)
    manifest.write(manifest_template.format(name=res.name, modules=modules,
                                            fetchto=dependencies))
    manifest.close()

    # syn Manifest
    maybe_mkdir(os.path.join("hdl", "syn"))
    maybe_mkdir(os.path.join("hdl", "syn", res.name))
    manifest_filename = os.path.join("hdl", "syn", res.name, "Manifest.py")
    manifest_template = open(
        os.path.join(data_dir, board, 'Manifest-syn.py')).read()
    print("Writing {}".format(manifest_filename))
    manifest = open(manifest_filename, "w")
    dependencies = compute_fetchto(manifest_filename)
    manifest.write(manifest_template.format(name=res.name,
                                            fetchto=dependencies))
    manifest.close()


    # ucf
    ucf_filename = os.path.join("hdl", "syn", res.name, res.name + ".ucf")
    ucf_template = open(os.path.join(data_dir, board, "top.ucf")).read()
    for k, v in slots.items():
        fmc_filename = os.path.join(
            data_dir, v, "{}-fmc{}.ucf".format(board, k))
        ucf_fmc = open(fmc_filename).read()
        ucf_template += ucf_fmc
    print("Writing {}".format(ucf_filename))
    ucf = open(ucf_filename, "w")
    ucf.write(ucf_template)
    ucf.close()

    # Copy tcl files
    for f in glob.glob(os.path.join(data_dir, board, "*.tcl")):
        print("Writing {}".format(os.path.basename(f)))
        shutil.copy(f, os.path.join("hdl", "syn", res.name))

def main():
    global data_dir

    aparser = argparse.ArgumentParser(description='wrtd config generator')
    aparser.add_argument('filename', help='YAML description file')
    args = aparser.parse_args()

    data_dir = os.path.dirname(__file__)

    res = Result()

    desc = parse(args.filename)
    res.name = desc['name']
    res.mt_config = build_mt_config(desc)
    slots = desc['slots']
    board = desc['board']
    if board not in boards:
        error('unknown board "{}"'.format(board))
    boards[board](res, desc, slots)

    generate_hdl(res, board, slots)

if __name__ == '__main__':
    main()
